@@ -1,4 +1,4 @@
-%% Copyright (C) 2024 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
+%% Copyright (C) 2024-2025 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
 %% Author: Vadim Yanitskiy <vyanitskiy@sysmocom.de>
 %%
 %% All Rights Reserved
@@ -40,7 +40,7 @@
          handle_call/3,
          handle_cast/2,
          terminate/2]).
--export([start_link/0,
+-export([start_link/1,
          process_pdu/2,
          fetch_erab/2,
          fetch_erab_list/1,
@@ -67,8 +67,14 @@
 -type enb_ue_id() :: 0..16#ffffff.
 -type erab_id() :: 0..16#ff.
 -type erab_uid() :: {mme_ue_id(), enb_ue_id(), erab_id()}.
+-type plmn_id() :: {MCC :: nonempty_string(),
+                    MNC :: nonempty_string()}.
 
--record(proxy_state, {erabs :: dict:dict(),
+-record(proxy_state, {conn_info :: sctp_server:conn_info(),
+                      erabs :: dict:dict(),
+                      enb_id :: undefined | non_neg_integer(),
+                      plmn_id :: undefined | plmn_id(),
+                      genb_id_str :: undefined | string(),
                       mme_ue_id :: undefined | mme_ue_id(),
                       enb_ue_id :: undefined | enb_ue_id(),
                       erab_id :: undefined | erab_id(),
@@ -85,9 +91,11 @@
 %% public API
 %% ------------------------------------------------------------------
 
--spec start_link() -> gen_server:start_ret().
-start_link() ->
-    gen_server:start_link(?MODULE, [], []).
+-spec start_link(ConnInfo) -> Result
+    when ConnInfo :: sctp_server:conn_info(),
+         Result :: gen_server:start_ret().
+start_link(ConnInfo) ->
+    gen_server:start_link(?MODULE, [ConnInfo], []).
 
 
 -type process_pdu_result() :: {proxy_action(), binary()}.
@@ -117,9 +125,10 @@ shutdown(Pid) ->
 %% gen_server API
 %% ------------------------------------------------------------------
 
-init([]) ->
+init([ConnInfo]) ->
     process_flag(trap_exit, true),
-    {ok, #proxy_state{erabs = dict:new(),
+    {ok, #proxy_state{conn_info = ConnInfo,
+                      erabs = dict:new(),
                       path = []}}.
 
 
@@ -167,6 +176,11 @@ terminate(Reason, #proxy_state{}) ->
 %% private API
 %% ------------------------------------------------------------------
 
+-spec set_logging_prefix(string()) -> ok.
+set_logging_prefix(Prefix) when is_list(Prefix) ->
+    logger:set_process_metadata(#{prefix => Prefix}).
+
+
 -spec erab_uid(erab_id(), proxy_state()) -> erab_uid().
 erab_uid(ERABId, #proxy_state{mme_ue_id = MmeUeId,
                               enb_ue_id = EnbUeId}) ->
@@ -195,6 +209,84 @@ erab_for_each(_UID, _Fun, []) ->
     ok.
 
 
+%% Parse PLMN-ID as per 3GPP TS 24.008, Figure 10.5.13
+%% | MCC digit 2 | MCC digit 1 |  octet 1
+%% | MNC digit 3 | MCC digit 3 |  octet 2
+%% | MNC digit 2 | MNC digit 1 |  octet 3
+-spec parse_plmn_id(<< _:24 >>) -> plmn_id().
+parse_plmn_id(<< MCC2:4, MCC1:4,
+                 MNC3:4, MCC3:4,
+                 MNC2:4, MNC1:4 >>) ->
+    MCC = parse_mcc_mnc(MCC1, MCC2, MCC3),
+    MNC = parse_mcc_mnc(MNC1, MNC2, MNC3),
+    {MCC, MNC}.
+
+
+-define(UNHEX(H), H + 48).
+
+parse_mcc_mnc(D1, D2, 16#f) ->
+    [?UNHEX(D1), ?UNHEX(D2)];
+
+parse_mcc_mnc(D1, D2, D3) ->
+    [?UNHEX(D1), ?UNHEX(D2), ?UNHEX(D3)].
+
+
+-spec parse_enb_id(tuple()) -> non_neg_integer().
+parse_enb_id({'macroENB-ID', << ID:20 >>}) -> ID;
+parse_enb_id({'homeENB-ID', << ID:28 >>}) -> ID;
+parse_enb_id({'short-macroENB-ID', << ID:18 >>}) -> ID;
+parse_enb_id({'long-macroENB-ID', << ID:21 >>}) -> ID.
+
+
+-spec genb_id_str(proxy_state()) -> string().
+genb_id_str(#proxy_state{plmn_id = {MCC, MNC},
+                         enb_id = ENBId}) ->
+    MCC ++ "-" ++ MNC ++ "-" ++ integer_to_list(ENBId).
+
+
+%% register a single per-eNB counter
+-spec ctr_reg(C, GlobalENBId) -> C
+    when C :: [ctr | _],
+         GlobalENBId :: string().
+ctr_reg([ctr, s1ap, proxy | _] = C0, GlobalENBId) ->
+    C1 = s1gw_metrics:enb_metric(C0, GlobalENBId),
+    %% counter may already exist, so catch exceptions here
+    catch exometer:new(C1, counter),
+    C0;
+
+ctr_reg(C0, _GlobalENBId) -> C0.
+
+
+%% register all per-eNB counters
+-spec ctr_reg_all(string()) -> list().
+ctr_reg_all(GlobalENBId) ->
+    Ctrs = s1gw_metrics:ctr_list(),
+    lists:map(fun(Name) -> ctr_reg(Name, GlobalENBId) end, Ctrs).
+
+
+%% increment the global and/or per-eNB counters
+-spec ctr_inc(C0, S) -> term()
+    when C0 :: [ctr | _],
+         S :: proxy_state().
+ctr_inc([ctr | _] = C0, S) ->
+    case S#proxy_state.genb_id_str of
+        undefined ->
+            %% increment the global counter only
+            s1gw_metrics:ctr_inc(C0);
+        GlobalENBId ->
+            %% increment both the global and per-eNB counters
+            C1 = s1gw_metrics:enb_metric(C0, GlobalENBId),
+            s1gw_metrics:ctr_inc(C0),
+            s1gw_metrics:ctr_inc(C1)
+    end.
+
+
+-spec tla_str(binary()) -> string().
+tla_str(TLA0) ->
+    TLA1 = list_to_tuple(binary_to_list(TLA0)),
+    inet:ntoa(TLA1).
+
+
 %% Encode an S1AP PDU
 -spec encode_pdu(s1ap_pdu()) -> {ok, binary()} |
                                 {error, {asn1, tuple()}}.
@@ -212,7 +304,7 @@ decode_pdu(Data) ->
 %% Process an S1AP PDU (binary)
 -spec handle_pdu_bin(binary(), proxy_state()) -> {process_pdu_result(), proxy_state()}.
 handle_pdu_bin(OrigData, S0) ->
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ALL),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ALL, S0),
     try decode_pdu(OrigData) of
         {ok, PDU} ->
             ?LOG_DEBUG("Rx S1AP PDU: ~p", [PDU]),
@@ -222,36 +314,36 @@ handle_pdu_bin(OrigData, S0) ->
                     ?LOG_DEBUG("Tx (~p) S1AP PDU: ~p", [Action, NewPDU]),
                     case Action of
                         forward ->
-                            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_ALL),
-                            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_PROC);
+                            ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_ALL, S1),
+                            ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_PROC, S1);
                         reply ->
-                            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ALL)
+                            ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ALL, S1)
                     end,
                     {{Action, NewData}, S1};
                 {forward, S1} ->
                     ?LOG_DEBUG("Tx (forward) S1AP PDU unmodified"),
-                    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_ALL),
-                    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED),
+                    ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_ALL, S1),
+                    ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED, S1),
                     {{forward, OrigData}, S1}
             catch
                 Exception:Reason:StackTrace ->
                     ?LOG_ERROR("An exception occurred: ~p, ~p, ~p", [Exception, Reason, StackTrace]),
-                    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_EXCEPTION),
-                    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
-                    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED),
+                    ctr_inc(?S1GW_CTR_S1AP_PROXY_EXCEPTION, S0),
+                    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S0),
+                    ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED, S0),
                     {{forward, OrigData}, S0} %% XXX: proxy as-is or drop?
             end;
         {error, {asn1, Error}} ->
             ?LOG_ERROR("S1AP PDU decoding failed: ~p", [Error]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_DECODE_ERROR),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_DECODE_ERROR, S0),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED, S0),
             {{forward, OrigData}, S0} %% XXX: forward as-is or drop?
     catch
         Exception:Reason:StackTrace ->
             ?LOG_ERROR("An exception occurred: ~p, ~p, ~p", [Exception, Reason, StackTrace]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_EXCEPTION),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_DECODE_ERROR),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_EXCEPTION, S0),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_DECODE_ERROR, S0),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED, S0),
             {{forward, OrigData}, S0} %% XXX: proxy as-is or drop?
     end.
 
@@ -260,12 +352,32 @@ handle_pdu_bin(OrigData, S0) ->
 -spec handle_pdu(s1ap_pdu(), proxy_state()) -> {{proxy_action(), s1ap_pdu()}, proxy_state()} |
                                                {forward, proxy_state()}.
 
+%% 9.1.8.4 S1 SETUP REQUEST
+handle_pdu({initiatingMessage,
+            #'InitiatingMessage'{procedureCode = ?'id-S1Setup',
+                                 value = C0}}, S0) ->
+    ?LOG_DEBUG("Processing S1 SETUP REQUEST"),
+    %% there's nothing to patch in this PDU, so we forward it as-is
+    %% TODO: check result of handle_ies(), inc. ?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR
+    {_, S1} = handle_ies(?'id-Global-ENB-ID',
+                         C0#'S1SetupRequest'.protocolIEs, S0),
+    {forward, S1};
+
+%% 9.1.8.5 S1 SETUP RESPONSE
+handle_pdu({successfulOutcome,
+            #'SuccessfulOutcome'{procedureCode = ?'id-S1Setup'}}, S) ->
+    ?LOG_DEBUG("Processing S1 SETUP RESPONSE"),
+    gtpu_kpi_enb_register(S),
+    %% there's nothing to patch in this PDU, so we forward it as-is
+    {forward, S};
+
+
 %% 9.1.3.1 E-RAB SETUP REQUEST
 handle_pdu({Outcome = initiatingMessage,
             #'InitiatingMessage'{procedureCode = ?'id-E-RABSetup',
                                  value = C0} = Msg}, S0) ->
     ?LOG_DEBUG("Processing E-RAB SETUP REQUEST"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_SETUP_REQ),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_SETUP_REQ, S0),
     case handle_ies(?'id-E-RABToBeSetupListBearerSUReq',
                     C0#'E-RABSetupRequest'.protocolIEs, S0) of
         {{ok, IEs}, S1} ->
@@ -274,9 +386,9 @@ handle_pdu({Outcome = initiatingMessage,
             {{forward, PDU}, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process E-RAB SETUP REQUEST: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             PDU = build_erab_setup_response_failure(S1),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ERAB_SETUP_RSP),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ERAB_SETUP_RSP, S1),
             {{reply, PDU}, S1} %% reply PDU back to sender
     end;
 
@@ -285,7 +397,7 @@ handle_pdu({Outcome = successfulOutcome,
             #'SuccessfulOutcome'{procedureCode = ?'id-E-RABSetup',
                                  value = C0} = Msg}, S0) ->
     ?LOG_DEBUG("Processing E-RAB SETUP RESPONSE"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_SETUP_RSP),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_SETUP_RSP, S0),
     case handle_ies(?'id-E-RABSetupListBearerSURes',
                     C0#'E-RABSetupResponse'.protocolIEs, S0) of
         {{ok, IEs}, S1} ->
@@ -294,7 +406,7 @@ handle_pdu({Outcome = successfulOutcome,
             {{forward, PDU}, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process E-RAB SETUP RESPONSE: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
@@ -303,7 +415,7 @@ handle_pdu({Outcome = initiatingMessage,
             #'InitiatingMessage'{procedureCode = ?'id-E-RABModify',
                                  value = C0} = Msg}, S0) ->
     ?LOG_DEBUG("Processing E-RAB MODIFY REQUEST"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MODIFY_REQ),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MODIFY_REQ, S0),
     case handle_ies(?'id-E-RABToBeModifiedListBearerModReq',
                     C0#'E-RABModifyRequest'.protocolIEs, S0) of
         {{ok, IEs}, S1} ->
@@ -312,7 +424,7 @@ handle_pdu({Outcome = initiatingMessage,
             {{forward, PDU}, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process E-RAB MODIFY REQUEST: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
@@ -321,7 +433,7 @@ handle_pdu({successfulOutcome,
             #'SuccessfulOutcome'{procedureCode = ?'id-E-RABModify',
                                  value = C0}}, S0) ->
     ?LOG_DEBUG("Processing E-RAB MODIFY RESPONSE"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MODIFY_RSP),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MODIFY_RSP, S0),
     %% there's nothing to patch in this PDU, so we forward it as-is
     %% TODO: check result of handle_ies(), inc. ?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR
     {_, S1} = handle_ies(?'id-E-RABModifyListBearerModRes',
@@ -331,56 +443,53 @@ handle_pdu({successfulOutcome,
     {forward, S2};
 
 %% 9.1.3.5 E-RAB RELEASE COMMAND
-handle_pdu({Outcome = initiatingMessage,
+handle_pdu({initiatingMessage,
             #'InitiatingMessage'{procedureCode = ?'id-E-RABRelease',
-                                 value = C0} = Msg}, S0) ->
+                                 value = C0}}, S0) ->
     ?LOG_DEBUG("Processing E-RAB RELEASE COMMAND"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_RELEASE_CMD),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_RELEASE_CMD, S0),
     case handle_ies(?'id-E-RABToBeReleasedList',
                     C0#'E-RABReleaseCommand'.protocolIEs, S0) of
-        {{ok, IEs}, S1} ->
-            C1 = C0#'E-RABReleaseCommand'{protocolIEs = IEs},
-            PDU = {Outcome, Msg#'InitiatingMessage'{value = C1}},
-            {{forward, PDU}, S1}; %% forward patched PDU
+        {{ok, _}, S1} ->
+            %% there's nothing to patch in this PDU, so we forward it as-is
+            {forward, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process E-RAB RELEASE COMMAND: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
 %% 9.1.3.6 E-RAB RELEASE RESPONSE
-handle_pdu({Outcome = successfulOutcome,
+handle_pdu({successfulOutcome,
             #'SuccessfulOutcome'{procedureCode = ?'id-E-RABRelease',
-                                 value = C0} = Msg}, S0) ->
+                                 value = C0}}, S0) ->
     ?LOG_DEBUG("Processing E-RAB RELEASE RESPONSE"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_RELEASE_RSP),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_RELEASE_RSP, S0),
     case handle_ies(?'id-E-RABReleaseListBearerRelComp',
                     C0#'E-RABReleaseResponse'.protocolIEs, S0) of
-        {{ok, IEs}, S1} ->
-            C1 = C0#'E-RABReleaseResponse'{protocolIEs = IEs},
-            PDU = {Outcome, Msg#'SuccessfulOutcome'{value = C1}},
-            {{forward, PDU}, S1}; %% forward patched PDU
+        {{ok, _}, S1} ->
+            %% there's nothing to patch in this PDU, so we forward it as-is
+            {forward, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process E-RAB RELEASE RESPONSE: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
 %% 9.1.3.7 E-RAB RELEASE INDICATION
-handle_pdu({Outcome = initiatingMessage,
+handle_pdu({initiatingMessage,
             #'InitiatingMessage'{procedureCode = ?'id-E-RABReleaseIndication',
-                                 value = C0} = Msg}, S0) ->
+                                 value = C0}}, S0) ->
     ?LOG_DEBUG("Processing E-RAB RELEASE INDICATION"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_RELEASE_IND),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_RELEASE_IND, S0),
     case handle_ies(?'id-E-RABReleasedList',
                     C0#'E-RABReleaseIndication'.protocolIEs, S0) of
-        {{ok, IEs}, S1} ->
-            C1 = C0#'E-RABReleaseIndication'{protocolIEs = IEs},
-            PDU = {Outcome, Msg#'InitiatingMessage'{value = C1}},
-            {{forward, PDU}, S1}; %% forward patched PDU
+        {{ok, _}, S1} ->
+            %% there's nothing to patch in this PDU, so we forward it as-is
+            {forward, S1};
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process E-RAB RELEASE INDICATION: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
@@ -389,7 +498,7 @@ handle_pdu({Outcome = initiatingMessage,
             #'InitiatingMessage'{procedureCode = ?'id-E-RABModificationIndication',
                                  value = C0} = Msg}, S0) ->
     ?LOG_DEBUG("Processing E-RAB MODIFICATION INDICATION"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MOD_IND),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MOD_IND, S0),
     IEs0 = C0#'E-RABModificationIndication'.protocolIEs,
     %% E-RAB to be Modified List
     %% TODO: handle {error, Reason}
@@ -406,7 +515,7 @@ handle_pdu({successfulOutcome,
             #'SuccessfulOutcome'{procedureCode = ?'id-E-RABModificationIndication',
                                  value = C0}}, S0) ->
     ?LOG_DEBUG("Processing E-RAB MODIFICATION CONFIRM"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MOD_CNF),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_ERAB_MOD_CNF, S0),
     IEs = C0#'E-RABModificationConfirm'.protocolIEs,
     %% E-RAB Modify List
     %% TODO: handle {error, Reason}
@@ -425,7 +534,7 @@ handle_pdu({Outcome = initiatingMessage,
             #'InitiatingMessage'{procedureCode = ?'id-InitialContextSetup',
                                  value = C0} = Msg}, S0) ->
     ?LOG_DEBUG("Processing INITIAL CONTEXT SETUP REQUEST"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_INIT_CTX_REQ),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_INIT_CTX_REQ, S0),
     case handle_ies(?'id-E-RABToBeSetupListCtxtSUReq',
                     C0#'InitialContextSetupRequest'.protocolIEs, S0) of
         {{ok, IEs}, S1} ->
@@ -434,7 +543,7 @@ handle_pdu({Outcome = initiatingMessage,
             {{forward, PDU}, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process INITIAL CONTEXT SETUP REQUEST: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
@@ -443,7 +552,7 @@ handle_pdu({Outcome = successfulOutcome,
             #'SuccessfulOutcome'{procedureCode = ?'id-InitialContextSetup',
                                  value = C0} = Msg}, S0) ->
     ?LOG_DEBUG("Processing INITIAL CONTEXT SETUP RESPONSE"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_INIT_CTX_RSP),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_INIT_CTX_RSP, S0),
     %% TODO: handle optional E-RAB Failed to Setup List IE
     case handle_ies(?'id-E-RABSetupListCtxtSURes',
                     C0#'InitialContextSetupResponse'.protocolIEs, S0) of
@@ -453,7 +562,7 @@ handle_pdu({Outcome = successfulOutcome,
             {{forward, PDU}, S1}; %% forward patched PDU
         {{error, Reason}, S1} ->
             ?LOG_NOTICE("Failed to process INITIAL CONTEXT SETUP RESPONSE: ~p", [Reason]),
-            s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR),
+            ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_PROC_ERROR, S1),
             {forward, S1} %% XXX: forward as-is or drop?
     end;
 
@@ -463,7 +572,7 @@ handle_pdu({initiatingMessage,
                                  value = #'UEContextReleaseRequest'{protocolIEs = IEs}}},
            #proxy_state{erabs = ERABs} = S) ->
     ?LOG_DEBUG("Processing UE CONTEXT RELEASE REQUEST"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_RELEASE_CTX_REQ),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_RELEASE_CTX_REQ, S),
     %% fetch {MME,eNB}-UE-S1AP-ID values (mandatory IEs)
     #'ProtocolIE-Field'{id = ?'id-MME-UE-S1AP-ID',
                         value = MMEUEId} = lists:nth(1, IEs),
@@ -481,7 +590,7 @@ handle_pdu({initiatingMessage,
                                  value = #'UEContextReleaseCommand'{protocolIEs = IEs}}},
            #proxy_state{erabs = ERABs} = S) ->
     ?LOG_DEBUG("Processing UE CONTEXT RELEASE COMMAND"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_RELEASE_CTX_CMD),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_RELEASE_CTX_CMD, S),
     #'ProtocolIE-Field'{id = ?'id-UE-S1AP-IDs',
                         value = S1APIDs} = lists:nth(1, IEs),
     case S1APIDs of
@@ -505,7 +614,7 @@ handle_pdu({successfulOutcome,
                                  value = #'UEContextReleaseComplete'{protocolIEs = IEs}}},
            #proxy_state{erabs = ERABs} = S) ->
     ?LOG_DEBUG("Processing UE CONTEXT RELEASE COMPLETE"),
-    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_RELEASE_CTX_COMPL),
+    ctr_inc(?S1GW_CTR_S1AP_PROXY_IN_PKT_RELEASE_CTX_COMPL, S),
     %% fetch {MME,eNB}-UE-S1AP-ID values (mandatory IEs)
     #'ProtocolIE-Field'{id = ?'id-MME-UE-S1AP-ID',
                         value = MMEUEId} = lists:nth(1, IEs),
@@ -538,6 +647,31 @@ handle_pdu(_PDU, S) ->
          Result :: {handle_ie_result(),
                     proxy_state()}.
 
+handle_ie([?'id-Global-ENB-ID'],
+          #'Global-ENB-ID'{'pLMNidentity' = PLMNId,
+                           'eNB-ID' = ENBId} = C, S0) ->
+    %% store PLMNId/ENBId
+    S1 = S0#proxy_state{plmn_id = parse_plmn_id(PLMNId),
+                        enb_id = parse_enb_id(ENBId)},
+    ?LOG_INFO("Global-ENB-ID: PLMN-ID=~p, eNB-ID=~p",
+              [S1#proxy_state.plmn_id,
+               S1#proxy_state.enb_id]),
+    %% use that as a context for logging
+    GlobalENBId = genb_id_str(S1),
+    set_logging_prefix("eNB " ++ GlobalENBId),
+    %% register per-eNB metrics
+    ctr_reg_all(GlobalENBId),
+    %% increment per-eNB ?S1GW_CTR_S1AP_PROXY_IN_PKT_ALL, if needed
+    %% this is needed to count S1 Setup Req correctly
+    S2 = case S0#proxy_state.genb_id_str of
+        GlobalENBId -> S1;
+        _ ->
+            Ctr = s1gw_metrics:enb_metric(?S1GW_CTR_S1AP_PROXY_IN_PKT_ALL, GlobalENBId),
+            s1gw_metrics:ctr_inc(Ctr),
+            S1#proxy_state{genb_id_str = GlobalENBId}
+    end,
+    {{ok, C}, S2};
+
 %% E-RAB SETUP REQUEST related IEs
 handle_ie([?'id-E-RABToBeSetupListBearerSUReq'], C, S) ->
     %% This IE contains a list of BearerSUReq, so patch inner IEs
@@ -569,6 +703,8 @@ handle_ie([?'id-E-RABSetupItemBearerSURes',
           #'E-RABSetupItemBearerSURes'{'e-RAB-ID' = ERABId,
                                        'transportLayerAddress' = TLA_In,
                                        'gTP-TEID' = << TEID_In:32/big >>} = C0, S) ->
+    %% indicate eNB's address to the gtpu_kpi module
+    gtpu_kpi_enb_set_addr({s1ap, tla_str(TLA_In)}),
     %% poke E-RAB FSM
     case erab_fsm_find(ERABId, S) of
         {ok, Pid} ->
@@ -856,6 +992,8 @@ handle_ie([?'id-E-RABSetupItemCtxtSURes',
           #'E-RABSetupItemCtxtSURes'{'e-RAB-ID' = ERABId,
                                      'transportLayerAddress' = TLA_In,
                                      'gTP-TEID' = << TEID_In:32/big >>} = C0, S) ->
+    %% indicate eNB's address to the gtpu_kpi module
+    gtpu_kpi_enb_set_addr({s1ap, tla_str(TLA_In)}),
     %% poke E-RAB FSM
     case erab_fsm_find(ERABId, S) of
         {ok, Pid} ->
@@ -963,4 +1101,38 @@ erab_fsm_find(RABId, #proxy_state{erabs = ERABs} = S) ->
     UID = erab_uid(RABId, S),
     dict:find(UID, ERABs).
 
+
+-spec gtpu_kpi_enb_register(proxy_state()) -> ok.
+gtpu_kpi_enb_register(#proxy_state{conn_info = ConnInfo} = S) ->
+    %% register eNB using its Global-eNB-ID
+    ok = gtpu_kpi:enb_register(genb_id_str(S)),
+    %% indicate UL/DL addresses
+    EnbAddr = inet:ntoa(maps:get(addr, ConnInfo)),
+    gtpu_kpi_enb_set_addr({sctp, EnbAddr}).
+
+
+-spec gtpu_kpi_enb_set_addr({Source, EnbAddr}) -> ok
+    when Source :: s1ap | sctp,
+         EnbAddr :: string().
+gtpu_kpi_enb_set_addr({Source, EnbAddr}) ->
+    gtpu_kpi_enb_set_addr(gtpu_kpi_ul_addr, {Source, ul, EnbAddr}),
+    gtpu_kpi_enb_set_addr(gtpu_kpi_dl_addr, {Source, dl, EnbAddr}),
+    ok.
+
+
+-spec gtpu_kpi_enb_set_addr(EnvParam, {Source, ULDL, EnbAddr}) -> ok
+    when EnvParam :: atom(),
+         Source :: s1ap | sctp,
+         ULDL :: ul | dl,
+         EnbAddr :: string().
+gtpu_kpi_enb_set_addr(EnvParam, {Source, ULDL, EnbAddr}) ->
+    case osmo_s1gw:get_env(EnvParam, s1ap) of
+        Source ->
+            ?LOG_DEBUG("GTP-U KPI ~p address ~p learned from ~p",
+                       [ULDL, EnbAddr, Source]),
+            gtpu_kpi:enb_set_addr({ULDL, EnbAddr});
+        Mode ->
+            ?LOG_DEBUG("GTP-U KPI ~p address mode ~p != ~p", [ULDL, Mode, Source])
+    end.
+
 %% vim:set ts=4 sw=4 et:
